// Copyright (C) 1999-2000 Logi Ragnarsson

package org.logi.crypto.io;
import org.logi.crypto.*;
import org.logi.crypto.keys.*;
import org.logi.crypto.sign.*;

import java.io.*;

/**
 * This InputStream reads the stream generated by a SignStream and
 * verifies the embedded signatures or hashes. It will not pass any
 * data through until that data has been verified.
 *
 * @see org.logi.crypto.io.SignStream
 * @version 1.0.6
 * 
 * @author <a href="http://www.logi.org/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@logi.org">logir@logi.org</a>) */
public class VerifyStream extends FilterInputStream{
    
    private HashState fs;
    private SignatureKey key;
    
    // Iff in is an instance of DecryptStream, then drainer will
    // reference it as well.  this is used to drain random garbage
    // inserted into the underlying stream when it is flushed.
    private DecryptStream drainer;

    /**
     * <code>buffer[bufPos..bufPos-1]</code> is the data which
     * has been read from the underlying stream but not from this. */
    private byte[] buffer;
    private int bufPos;
    private int bufSize;

    /**
     * Create a new VerifyStream object. It matches the SignStream object
     * with the same <code>blockSize</code> and <code>fs</code> and with the
     * key that matches <code>key</code>
     *
     * @see org.logi.crypto.io.SignStream
     */
    public VerifyStream(InputStream in, int blockSize, SignatureKey key, HashState fs) {
        super(in);
	if(in instanceof DecryptStream)
	    drainer = (DecryptStream)in;
        this.key = key;
        this.bufSize=(blockSize/fs.blockSize())*fs.blockSize();
        if (this.bufSize==0)
            this.bufSize=fs.blockSize();
        buffer=new byte[bufSize];
        bufPos=bufSize; // no unread data in buffer
        this.fs = fs;
    }
    
    /**
     * Reads the next byte of data from this input stream. The value
     * byte is returned as an int in the range 0 to 255. If no byte is
     * available because the end of the stream has been reached, the
     * value -1 is returned. This method blocks until input data is
     * available, the end of the stream is detected, or an exception
     * is thrown.  */
    public synchronized int read() throws IOException{
        if (bufPos==bufSize)
            fillBuffer(true);
        if(buffer==null)
            return -1;
        return buffer[bufPos++] & 0xff;
    }
    
    /**
     * If the size of the next buffer has been read, but the buffer itself
     * hasn't been read yet, then newBufSize is that size. Otherwise it
     * will be -1. */
    private int newBufSize=-1;
   
    /**
     * Fills the buffer with data and verifies the signature or
     * fingerprint.  Sets bufSize==0 if EOF is reached. Drains random
     * garbage. If mayBlock is false, the method returns without doing
     * anything if it would otherwise block.
     */
    private void fillBuffer(boolean mayBlock) throws IOException {
       try {
	  if((!mayBlock) && in.available()<4)
	    // We can't read the buffer-size and musn't block.
	    return;
	  
	  if(newBufSize==-1) {
	     newBufSize = Crypto.readInt(in);
	     if(newBufSize>buffer.length)
	       throw new IOException("The stream contains a block of size "+newBufSize+" which is larger than the buffer size "+buffer.length);
	  }
	  
	  if((!mayBlock) && in.available()<newBufSize)
	    // We can't read the buffer and musn't block.
	    return;
       } catch (IOException ioe) {
	  buffer=null;   // There are no more blocks to read. Close buffer, but don't throw the exception.
       }
       
       // Now, the buffer size is read and OK and we know we can fill the
       // buffer without illegal blocking, so we can safely set bufSize
       // and newBufSize to the values they should have after the buffer
       // is filled.
       	   
       bufSize = newBufSize;
       newBufSize = -1;
       
       try {
	  bufPos=0;
	  while(bufPos<bufSize){
	     int n=in.read(buffer,bufPos,bufSize-bufPos);
	     if (n<0){
		// EOF inside block! terminate!!
		buffer=null;
		return;
	     }
	     bufPos+=n;
	  }
	  // buffer is filled.
	    
	  bufPos =0;
	  byte[] rfp = new byte[ (key==null) ? fs.hashSize() : key.signatureSize() ];
	  int rfpPos=Crypto.readBlock(in, rfp, 0, rfp.length);
	  if (rfpPos<0){
	     // EOF inside signature! terminate!!
	     buffer=null;
	     return;
	  }
	  // rfp contains the next fingerprint or signature from the
	  // stream. buffer[0..bufSize-1] contains the undelivered data.
	    
	  fs.update(buffer,0,bufSize);
	  Fingerprint fp=fs.calculate();
	  if(key==null){
	     if(!Crypto.equal(rfp,fp.getBytes()))
	       throw new IOException("Data was not verified");
	  } else {
	     Signature sig = new Signature(fp.getHashFunc(), rfp);
	     try{
		if(!key.verify(sig,fp))
		  throw new IOException("Signature was not verified");
	     } catch (KeyException e) {
		throw new IOException(e.getMessage());
	     }
	  }
	  // Signature/fingerprint is verified
       } catch (IOException ioe) {
	  buffer=null;   // We have unverified data in the buffer, close stream.
	  throw ioe;
       }
    }
    
    /**
     * Reads up to len bytes of data from this input stream into an array of
     * bytes. This method blocks until some input is available.
     * <p>
     * The actual number of bytes read is returned or -1 if the end of the
     * stream is reached.
     */
    public synchronized int read(byte b[], int off, int len) throws IOException{
        if(bufPos==bufSize)
            fillBuffer(true);
        if(buffer==null)
            return -1;
        
        if (len>bufSize-bufPos)
            len=bufSize-bufPos;
        System.arraycopy(buffer,bufPos, b,off, len);
        bufPos+=len;
        return len;
    }
    
    /**
     * Drain the underlying stream if it is a DecryptStream.
     *
     * @see org.logi.crypto.io.DecryptStream#drain
     * */
    public int drain(){
	if(drainer!=null)
	    return drainer.drain();
	else return 0;
    }
    
    /** 
     * Returns the number of bytes that can be read from this input stream
     * without blocking. */
    public int available() throws IOException{
       if(bufSize==bufPos)
	 fillBuffer(false);
       return bufSize-bufPos;
    }
    
    /**
     * Does nothing and returns 0.
     */
    public long skip(long n) throws IOException {
	return 0;
    }
   
    /** 
     * Returns false. (This could be implemented, but I've never seen it
     * used. Mail me if you want it!) */
    public boolean markSupported(){
        return false;
    }
    
}
